FAIL_MESSAGE = "FAIL"
WARNING_MESSAGE = "WARNING"
# Errors for Hansel
MISSING_TILES_ERROR_1 = "Missing Tiles Error 1"
MIXED_SAMPLE_ERROR_2 = "Mixed Sample Error 2"
AMBIGUOUS_RESULTS_ERROR_3 = "Ambiguous Results Error 3"
NON_CONFIDENT_RESULTS_ERROR_4 = "Non Confident Results Error 4"
INTERMEDIATE_SUBTYPE_WARNING = "Intermediate Subtype Warning"
